from .shap_model_describer import SHAPModelDescriber  # TODO: fix
from .enums import Constituency, Aspect
from .describer import Describer


class Describable:
    """
    A *Trustworthy-AI* interface that defines the capabilities of objects (e.g., `Models`, `Graphs`)
    that are `Describable, i.e., they can self-describe their properties and behaviors. For example,
    a `Describable` `Model` should be able to report the data that was used to train it, provide an
    importance-ranked list of its input features on a global basis, etc.
    """

    @property
    def description(self):
        return getattr(self, "__description", {})

    @description.setter
    def description(self, value):
        setattr(self, "__description", value)

    def describe(self, constituency=Constituency.ANY, aspect=Aspect.ANY):
        """
        Returns a description of the model's behavior and properties based on `Who's asking` for `what`.

            Parameters:
                constituent : Constituency: The Constituency asking for the explanation `Who`
                aspect : The Aspect of the question. `What`

            Returns:
                out : Description of Model's behavior and properties
        """
        describer = Describer(self)
        describer.shap_describer = SHAPModelDescriber(self._native_model, self.prepared_data)
        describer.generate_report(constituency, aspect)
        return describer



from typing import Dict, NoReturn, List
import pandas as pd
from h1st.model.predictive_model import PredictiveModel
from .oracle import Oracle
from .student import RandomForestModeler, AdaBoostModeler
from .ensemble import Ensemble


class TimeSeriesOracle(Oracle):
    def __init__(self, teacher: PredictiveModel,
                 students,
                 ensembler):
        super().__init__(teacher, students, ensembler)
        self.stats = {}

    @classmethod
    def generate_features(cls, data: Dict):
        '''
        Generate features to train the Student model.
        By default, we flatten all data points of the grouped dataframe.
        Overwrite this method to do custom featurization work.
        :param data: a dictionary of a grouped dataframe.
        '''
        df = data['data']
        ret = pd.DataFrame(df.values.reshape(1, df.shape[0] * df.shape[1]))
        return {'data': ret}

    @classmethod
    def generate_data(cls, data: Dict, teacher: PredictiveModel, stats: Dict) -> Dict:
        '''
        Generate data to train the Student model
        :param data: unlabeled data in form of {'X': pd.DataFrame}
        :returns: a dictionary of features and teacher's prediction.
        '''
        if 'X' not in data:
            raise ValueError('Please provide data in form of {\'X\': pd.DataFrame}')

        df = data['X']

        id_col = stats['id_col']
        ts_col = stats['ts_col']
        features = stats['features']

        if id_col is not None and id_col not in df.columns:
            raise ValueError(f'{id_col} does not exist')

        if ts_col is not None and ts_col not in df.columns:
            raise ValueError(f'{ts_col} does not exist')

        if id_col is not None or ts_col is not None:
            groupby_cols = []
            if id_col is not None:
                groupby_cols.append(id_col)
            if ts_col is not None:
                groupby_cols.append(ts_col)

            features_list = []
            teacher_preds = []
            for _, group_df in df.groupby(groupby_cols):
                group_df.drop(groupby_cols, axis=1, inplace=True)
                if features is not None:
                    group_df = group_df[features]

                teacher_pred = teacher.predict({'X': group_df})
                if 'predictions' not in teacher_pred:
                    raise KeyError('Teacher\'s output must contain a key named `predictions`')

                teacher_preds.append(teacher_pred['predictions'])

                features_list.append(cls.generate_features({'data': group_df})['data'])
            df_features = pd.concat(features_list).fillna(0)
        else:
            if features is not None:
                df = df[features]
            teacher_preds = teacher.predict({'X': df})['predictions']
            df_features = cls.generate_features({'data': df})

        return {'X': df_features, 'y': pd.Series(teacher_preds)}

    def predict(self, input_data: Dict) -> Dict:
        '''
        Implement logic to generate prediction from data. The TimeSeries Oracle expects the same `id_col`, `ts_col` and features provided during `build` phase to be in the provided data. It automatically process the data the same way to that of the `build` phase.
        :params input_data: an dictionary with key `X` containing the data to get predictions.
        :returns: a dictionary with key `predictions` containing the predictions
        """
        '''
        return super().predict(input_data)



import sys
from types import SimpleNamespace
from typing import List, Union, Any, NoReturn, Dict

from .h1step import Node, Action
from .h1step_containable import NodeContainable
from h1st.exceptions.exception import GraphException
from h1st.core.viz import DotGraphVisualizer
from h1st.trust.trustable import Trustable


# __TODO__: refactor to Flows and Steps

class Graph(NodeContainable, Trustable):
    """
    A Graph is itself a NodeContainable, meaning it can be enclosed within a Node,
    forming a hierarchy of Graphs
    """

    def __init__(self, node_validation_schema_name='NODE_VALIDATION_SCHEMA'):
        """
        :param node_validation_schema_name: name of the variable in config.py to load schema definition fornode_validation_schema_name
        """
        super().__init__()
        self._node_validation_schema = None
        self._node_validation_schema_name = node_validation_schema_name

        self._nodes = SimpleNamespace()
        self._last_added_node = None

        # map { node's id: number} to trake the used ids.
        # with number=0 if id is manual provided, number=1 if id is generated
        self._used_node_ids = {}

    @property
    def nodes(self) -> SimpleNamespace:
        """
        Gets all nodes of the graph in unspecified orders
        Gets a specific node by ID: nodes.<ID>
        """
        return self._nodes

    def start(self) -> 'Graph':
        """
        Initial action to begin adding nodes to a (fresh) Graph. A node with the id='start' will be
        automatically added to the Graph.
        """
        return self.add(Action(id='start'))

    def add(self,
            node: Union[Node, NodeContainable, None] = None,
            yes: Union[Node, NodeContainable, None] = None,
            no: Union[Node, NodeContainable, None] = None,
            id: str = None
            ) -> Union[Node, List[Node]]:
        """
        Adds a new Node or NodeContainable to this graph. Period keeps a running preference to the current possition in the graph to be added
        If the object to be added is a NodeContainable then a new node will be automatically instanciated to contain that object and the node is added to this graph.
        The new node's id can be specified or automatically inferred from the NodeContainable's type.

        :param node: Node or NodeContainable object to be added to the graph
        :param yes/no: Node or NodeContable object to be added to the graph following a conditional (Decision) node
        :param from_: the node to which the new node will be connected
        :param id: the id of the new node

        :return:
            new added node if adding a single node
            Or new added node for yes branch if adding yes node only (without no node) following a condition node
            Or new added node for no branch if adding no node only (without yes node) following a condition node
            Or [new added node for yes branch, new added node for no branch] if adding both yes & no nodes following a condition node

        .. code-block:: python
            :caption: Cyber Security example to handle injection and replacement attacks

            import h1st.core as h1
            from h1st.core import NodeContainable, Decision

            class MyGraph(h1.Graph):
                def __init__(self):
                    super().__init__()

                    imc, rec = self
                        .start()
                        .add(GenerateWindowEvents())
                        .add(Decision(InjectionEventClassifier()))
                        .add(
                            yes=InjectionMessageClassifier(),
                            no=Decision(ReplacementEventClassifier())
                        )

                    rec.add(
                        yes=ReplacementMessageClassifier(),
                        no=BuildNormalResult()
                    )

                    self.end()

        """
        return self._add_and_connect(node, yes, no, id, self._last_added_node)

    def end(self) -> 'Graph':
        """
        This method is required after adding all nodes to the graph. The end node with id='end' will be automatically added to the graph.
        All leaf nodes (without outgoing edges) will be automatically connected to the end node.
        Consolidate ids for nodes using the same NodeContainable type without provided id. Ids will be Xyz1, Xyz2, ... with class Xyz inherits from NodeContainable
        """
        end_node = self.add(Action(id='end'))

        # connect all nodes without out-going edges to end node
        for id, node in self.nodes.__dict__.items():
            if not node.edges and id != end_node.id:
                self._connect_nodes(node, end_node)

        return end_node

    def execute(self, command: str, data: Union[Dict, List[Dict]]) -> Union[Dict, List[Dict]]:
        """
        The graph will scan through nodes to invoke appropriate node's function with name = value of command parameter.
        Everytime the graph invokes the appropreate function of the node, it will passing an accumulated dictionary as the input and merge result of the function into the accumulated dictionary.

        :param command: for Node or NodeContainable object to decide which function will be invoked during executing the graph
        :param data: input data to execute.
            if data is a dictionary, the graph will execute one.
            if data is a list of dictionary, the graph will execute multiple time

        :return:
            single dictionary if the input is a single dictionary
            Or list of dictionary if the input is a list of dictionary

        .. code-block:: python
            :caption: Example graph for Cyber Security and how to execute the graph

            import h1st.core as h1
            from h1st.core import NodeContainable, Decision

            class MyGraph(h1.Graph):
                def __init__(self):
                    imc, rec = self
                        .start()
                        .add(GenerateWindowEvents())
                        .add(Decision(InjectionEventClassifier()))
                        .add(
                            yes=InjectionMessageClassifier(),
                            no=Decision(ReplacementEventClassifier())
                        )

                    rec.add(
                        yes=ReplacementMessageClassifier(),
                        no=BuildNormalResult()
                    )

                    self.end()

            g = MyGraph()
            result = g.execute(command='predict', data={'df': my_dataframe})
        """
        if isinstance(data, list):
            return [self._execute_one(command, item) for item in data]

        return self._execute_one(command, data)

    def predict(self, data) -> Any:
        """ A shortcut function for the "execute" function with command="predict" """
        return self.execute('predict', data)

    def visualize(self):
        """Visualizes the flowchart for this graph"""
        vs = DotGraphVisualizer(self)
        return vs

    def describe(self):
        pass

    def explain(self):
        pass

    def _wrap_and_add(self, node: Union[Node, NodeContainable], id: str = None):
        """
        Wraps NodeContainable to a Node if needed. Override node's id if an id provided. Then adds the node to the graph.

        :param node: a Node or ContainableNode
        :param id: id for the node

        :return: the newly added node
        """
        if not isinstance(node, (Node, NodeContainable)):
            raise GraphException('object to add to a graph must be an instance of Node or NodeContainable')

        if isinstance(node, NodeContainable):
            containable = node
            node = Action(containable)

        id = id or node.id
        if id:  # manual provided id
            self._used_node_ids[id] = 0
        else:  # automatic id
            id = self._generate_id(node)
            self._used_node_ids[id] = 1

        node._id = id
        node.graph = self
        setattr(self.nodes, id, node)

        return node

    def _generate_id(self, node: Node):
        """
        Automatically generates an id for the node.
        If multiple instances of the same NodeContainable type are added,
        the ids of the node will be ClassName, ClassName2, ClassName3, etc
        """
        classname = (node._containable or node).__class__.__name__.split('.')[-1]

        if classname not in self._used_node_ids:
            return classname

        for i in range(2, sys.maxsize):
            id = f'{classname}{i}'
            if id not in self._used_node_ids:
                return id

    def _connect_nodes(self, from_: Node, to: Node, edge_label=None) -> NoReturn:
        """
        Connects "from_" node to the "to" node

        :param from_: the source node
        :param to: the destination node
        :edge_label: the label for the edge between from_ and to
        """
        if not from_:
            return

        if edge_label not in ['yes', 'no', None]:
            raise GraphException(
                f'edge_label="{edge_label}" is not supported')

        from_.edges.append(
            (to, edge_label)
        )

    def _execute_one(self, command: str, data: Dict) -> Dict:
        """
        Executes the graph exactly 1 time

        :param command: for Node or NodeContainable object to decide which function will be invoked during executing the graph
        :param data: input data to execute the graph

        :return: result as a dictionary
        """
        output = self.nodes.start._execute(command, data, {})

        if self.nodes.end.transform_output:
            output = self.nodes.end.transform_output(output)

        return output

    def _add_and_connect(self,
                         node: Union[Node, NodeContainable, None] = None,
                         yes: Union[Node, NodeContainable, None] = None,
                         no: Union[Node, NodeContainable, None] = None,
                         id: str = None,
                         from_: Union[Node, None] = None
                         ) -> Union[Node, List[Node]]:
        """
        Adds node/yes/no nodes to self.nodes and connect from_node to newly added nodes
        """
        if id == 'start' and hasattr(self.nodes, 'start'):
            raise GraphException('Graph.start() may only be called once')

        if id == 'end' and hasattr(self.nodes, 'end'):
            raise GraphException('Graph.end() may only be called once')

        if hasattr(self.nodes, 'end'):
            raise GraphException('not allow to add a node after Graph.end()')

        if id and hasattr(self.nodes, id):
            raise GraphException(f'Node id={id} is duplicated')

        # connect to latest node if not adding from a node
        if not from_:
            from_ = self._last_added_node

        # add a single node
        if node:
            node = self._wrap_and_add(node, id)
            self._connect_nodes(from_, node)

            # keep reference to the latest node
            self._last_added_node = node

            return node

        # add nodes with edge_label 'yes' / 'no'
        return_nodes = []

        if yes:
            node = self._wrap_and_add(yes)
            self._connect_nodes(from_, node, 'yes')
            return_nodes.append(node)

        if no:
            node = self._wrap_and_add(no)
            self._connect_nodes(from_, node, 'no')
            return_nodes.append(node)

        # keep reference to the latest node
        self._last_added_node = return_nodes[0] if len(return_nodes) == 1 else None

        # chaining will return array if having more than one node, otherwise return single node
        return return_nodes[0] if len(return_nodes) == 1 else return_nodes



from typing import Any, List
from h1st.model.ml_model import MLModel

from h1st.model.predictive_model import PredictiveModel

class Ensemble(PredictiveModel):
    """
    Base ensemble class.
    """



class Explainer:
    """
    An Explainer is one that stores all relevant details about the explanation
    of a decision that was made by the Model or Graph. For example, details such as
    `what was the original decision`, `what were the key factors that led to that decision` and
    `what type of model was used to make the decision` etc.
    """

    def __init__(self, h1stmodel, decision):
        self.lime_explainer = None
        self.decision_description(decision, h1stmodel)

    def decision_description(self, decision, h1stmodel):
        _dict = {}
        model = h1stmodel._native_model
        _dict["model_name"] = str(type(model).__name__)
        _dict["data_set_name"] = h1stmodel.dataset_name
        _dict["data_set_description"] = h1stmodel.dataset_description
        _dict["label_column"] = h1stmodel.label_column
        _dict["decision_input"] = decision[0]
        _dict["decision"] = decision[1]
        self.decision_describer = _dict

    def generate_report(self, decision, constituency, aspect):
        pass



from copy import deepcopy
from typing import Dict, List, Tuple, Union
import logging

from configparser import ConfigParser
from itertools import combinations, product
import itertools
import numpy as np
import pandas as pd

from h1st.model.model import Model
from h1st.model.modeler import Modeler

logger = logging.getLogger()
logger.setLevel(logging.INFO)

CLASSES_DTYPES = [str, object, bool, pd.CategoricalDtype]


class CombinationSegmentor(Model):
    def __init__(self):
        super().__init__()

    def process(
        self, 
        input_data: Dict, 
        by: Dict, 
        segmentation_levels: List,
        min_data_size: int
    ) -> Tuple[Dict]:
        """
        Create all combination 
        :param input_data: Input data for training and inference.
        :param by: Utilize this value to segment input_data.
        :param segmentation_levels: List
            Number of features being used to create each segment. If level=1, then
            take a filter from only one featrue to create a segment. If level=2, then 
            take a filter from each feature and create a combined filter to create a 
            segment. To create a combined filter, do and operation between two filters.
            
            For example, given an above segmentation_features, if level=3, then you will
            have the following 4 segments.
                segment_1: (sepal_size, (None, 18.5)) & (sepal_aspect_ratio, (0, 0.65)) & (species, [0, 1])
                segment_2: (sepal_size, (None, 18.5)) & (sepal_aspect_ratio, (0.65, 1)) & (species, [0, 1])
                segment_3: (sepal_size, (18.5, None)) & (sepal_aspect_ratio, (0, 0.65)) & (species, [0, 1])
                segment_4: (sepal_size, (18.5, None)) & (sepal_aspect_ratio, (0.65, 1)) & (species, [0, 1])
            If level=1, then you will have the following 5 segments.
                segment_1: (sepal_aspect_ratio, (0, 0.65))
                segment_2: (sepal_aspect_ratio, (0.65, 1))
                segment_3: (sepal_size, (None, 18.5))
                segment_4: (sepal_size, (18.5, None))
                segment_5: (species, [0, 1])
        :param min_data_size: Minimum number of data points per segment. 
            If the number of data point in a segment is less than this value, 
            then, merge the small segment with other close segments.

        """
        segmentation_logics = self.create_filter_combinations(by, segmentation_levels)
        
        by_key = list(by.keys())
        if 'dataframe' in input_data:
            data = {}
            if 'X_train' in input_data['dataframe']:
                data['X'] = input_data['dataframe']['X_train']
                data['y'] = input_data['dataframe']['y_train']
            segmentation_results = self.get_segments_from_dataframe(
                data, segmentation_logics, by_key)
            if 'X_train' in input_data['dataframe']:
                for name, segmentation_result in segmentation_results.items():
                    segmentation_result['X_train'] = segmentation_result['X']
                    del segmentation_result['X']
                    segmentation_result['y_train'] = segmentation_result['y']
                    del segmentation_result['y']
        elif 'json' in input_data:
            segmentation_results = self.get_segments_from_json(input_data['json'], segmentation_logics)
        else:
            raise KeyError('input_data does not include "dataframe" or "json" key.')
        filtered_segmentation_results = self.filter_small_segments(segmentation_results, min_data_size)
        # Done: make segments from (X and Y) 
        # TODO: make get_segments work with JSON data
        # TODO: self.auto_merge_small_segments
        self.stats = {'segmentation_logics': segmentation_logics}
        return (filtered_segmentation_results, segmentation_logics)

    def auto_merge_small_segments(self, segments, min_data_size, filter_combinations):
        pass

    def create_feature_combination(self, list_of_filters, level):
        res = []
        for feature_combinaton in itertools.combinations(list_of_filters, level):
            res.append(feature_combinaton)
        return res 

    def create_combinations_with_levels(self, list_of_filters, levels):
        combinations_with_levels = []
        for level in levels:
            if len(list_of_filters) < level:
                raise ValueError(f'level ({level}) cannot be bigger than the number \
                                 of features ({len(list_of_filters)})')
            feature_combinations = self.create_feature_combination(list_of_filters, level)
            for feature_combination in feature_combinations:
                combinations = list(itertools.product(*feature_combination))
                combinations_with_levels.append((level, combinations))
        return combinations_with_levels

    def create_filter_combinations(self, by, levels=[1]):
        list_of_filters = []
        for feature, filters in by.items():
            filters_per_feature = []
            for filter in filters:
                filters_per_feature.append((feature, filter))
            list_of_filters.append(filters_per_feature)
        combinations_with_levels = self.create_combinations_with_levels(list_of_filters, levels)
        filter_combinations = {}
        idx = 0
        for level, combs in combinations_with_levels:
            for comb in combs:
                filter_combinations[f'segment_{idx}_lvl_{level}'] = list(comb)
                idx += 1
        return filter_combinations

    def get_filtered_indices_from_range(self, dataframe, feature, data_range):
        start, end = data_range
        if start and end:
            return (start <= dataframe[feature]) & (dataframe[feature] < end)
        if start:
            return (start <= dataframe[feature])
        if end:
            return (dataframe[feature] < end)
        return None

    def get_filtered_indices_from_set(self, dataframe, feature, data_set):
        return (dataframe[feature].isin(data_set))

    def get_filtered_indices(self, dataframe, feature, filter):
        if feature not in dataframe:
            raise KeyError(f'feature: {feature} is not in dataframe')
        if isinstance(filter, tuple):
            if len(filter) != 2:
                raise ValueError('Range filter should have only two values, start and end')
            filter = self.get_filtered_indices_from_range(dataframe, feature, filter)
        elif isinstance(filter, list):
            filter = self.get_filtered_indices_from_set(dataframe, feature, filter)
        else:
            raise ValueError(f'Provided filter should be either list or tuple, not {type(filter)}')    
        return filter.values

    def get_segments_from_dataframe(self, dataframe, segmentation_logics, segmentation_features_key):
        X = dataframe['X']
        X_features = list(X.columns)
        for item in segmentation_features_key: X_features.remove(item)

        segments = {}
        for segment_name, comb in segmentation_logics.items():
            list_filtered_indices = []
            for feature, filter in comb:
                filtered_indices = self.get_filtered_indices(X, feature, filter)
                list_filtered_indices.append(filtered_indices)
            segment_indices = np.column_stack(list_filtered_indices).all(axis=1)
            segments[segment_name] = {'X': X.loc[segment_indices, X_features]}
            if 'y' in dataframe:
                segments[segment_name]['y'] = dataframe['y'].loc[segment_indices]
        return segments

    def filter_small_segments(self, segments, min_data_size):
        res = {}
        for segment_name, segment in segments.items():
            if segment['X_train'].shape[0] < min_data_size:
                logger.info(f'{segment_name} has {segment["X_train"].shape[0]} number of data points,\
                            which is smaller than min_data_size: {min_data_size}')
                continue
            res[segment_name] = segment
        return res


class MaxSegmentationModel(Model):

    name = 'max_segmentation_model'

    def __init__(self):
        super().__init__()
        self.stats = {}

    def process(self, input_data: dict) -> dict:
        '''
        input dict requires key "X" contain a pandas dataframe. returns dict
        with keys "segment_info" containing dict of metadata about each segment
        and "segment_data" with a dict formatted as {segment_name: pd.DataFrame}
        '''
        X = input_data['X']

        segments = self.stats['segment_info'].keys()

        out_data = {}
        out_info = {}
        for name in segments:
            idx = self.get_segment_idx(X, name)
            if len(idx) == 0:
                continue

            out_data[name] = X.loc[idx].copy()
            out_info[name] = self.group_info(name)

        return {'segment_info': out_info,
                'segment_data': out_data}

    def group_info(self, name):
        return self.stats['segment_info'][name]

    def get_segment_idx(self, df: pd.DataFrame, segment_name: str):
        '''get indices of dataframe in segment'''
        grp = self.stats['segment_info'][segment_name]
        tmp = df.apply(lambda x: is_in_segment(x, grp), axis=1)
        return df[tmp].index.tolist()

    @classmethod
    def construct_model(cls, segments: dict):
        model = cls()
        model.stats['segment_info'] = segments
        return model


class MaxSegmentationModeler(Modeler):
    def __init__(
        self, 
        model_class = MaxSegmentationModel,
        config: Union[dict, str] = None,
    ):
        self.model_class = model_class
        if config:
            self._load_config(config)

    def _load_config(self, config: Union[dict,str]):
        if isinstance(config, str):
            self._load_config_file(config)
        elif isinstance(config, dict):
            self._load_config_dict(config)
        elif config is None:
            self.config = None
        else:
            raise TypeError('Config must be path to config file or dict of '
                            'segmentation features')

    def build_model(self, input_data: dict, config: Union[dict, str]=None):
        if config is not None:
            self._load_config(config)
            
        if self.config is None:
            raise ValueError('Must provide config or initialize with config')

        # combine X_train and X_test to create full dataframe to create groups
        # X_test is optional
        X = pd.concat([input_data['X_train'],
                       input_data.get('X_test')], axis=0)

        # Create bins for individual features
        self._create_bins(X)

        config = self.config
        min_pts = config['min_segment_size']
        seg_features = config['segmentation_features']
        max_levels = len(seg_features)

        segments = {}
        for level in np.arange(max_levels)+1:
            # Merge small segments in level 1 to make "other" segment,
            # not for level 2+
            if level == 1:
                merge_small = True
            else:
                merge_small = False

            grps = self._create_segments_for_level(X, level,
                                                   merge_small=merge_small)
            segments.update(grps)

        model = self.model_class.construct_model(segments)
        model.metrics = self.evaluate_model(input_data, model)
        return model

    def evaluate_model(self, input_data: dict, model: Model):
        X_train = input_data['X_train']
        X_test = input_data.get('X_test')
         
        train_segments = model.process({'X': X_train})['segment_data']
        if X_test is not None:
            test_segments = model.process({'X': X_test})['segment_data']
        else:
            test_segments = {}

        metrics = {}
        all_groups = set([*train_segments.keys(), *test_segments.keys()])
        for name in all_groups:
            train = train_segments.get(name)
            test = test_segments.get(name)

            if train is None:
                n_train = 0
            else:
                n_train = train.shape[0]

            if test is None:
                n_test = 0
            else:
                n_test = test.shape[0]

            metrics[name] = {'train_pts_in_segment': n_train,
                             'test_pts_in_segment': n_test}

        return metrics

    def _create_segments_for_level(self, labels: pd.DataFrame, level: int,
                                  merge_small=False):
        '''If merge_small is true then small groups are merged. Otherwise they
        are cut. For level>1, its probably best to cut them, for level==1 you
        should merge them for complete data coverage'''
        config = self.config
        min_pts = config['min_segment_size']
        seg_features = config['segmentation_features']

        features = [x['feature'] for x in seg_features]
        bins = [x['bins'] for x in seg_features]
        bin_types = [x['bin_type'] for x in seg_features]

        feat_grps = combinations(features, level)
        good_groups = []
        small_groups = []
        for fgrp in feat_grps:
            bin_grps = product(*[bins[features.index(x)] for x in fgrp])
            for bgrp in bin_grps:
                tmp = {
                    'features': list(fgrp),
                    'bins': list(bgrp),
                    'bin_types': [bin_types[features.index(x)] for x in fgrp],
                    'level': int(level),
                }
                in_grp = labels.apply(
                    lambda x: is_in_segment(x, tmp), axis=1
                )
                tmp['n_pts'] = sum(in_grp)
                if tmp['n_pts'] < min_pts:
                    small_groups.append(tmp)
                else:
                    good_groups.append(tmp)

        # Merge small groups
        if merge_small:
            other_group = self.merge_groups(small_groups)
            good_groups.extend(other_group)

        # Name groups
        out_segments = {}
        for i, grp in enumerate(good_groups):
            group_name = f'lvl_{level}-group_{i}'
            out_segments[group_name] = grp

        return out_segments

    def _create_bins(self, labels: pd.DataFrame):
        config = deepcopy(self.config)
        min_pts = config['min_segment_size']
        seg_features = deepcopy(config['segmentation_features'])

        for seg in seg_features:
            if 'bins' in seg.keys():
                continue

            feat = seg['feature']
            n_bins = seg.get('n_bins')
            bins = get_bins(labels[feat], n_bins=n_bins)
            good_bins = []
            small_bins = []
            for bin_ in bins:
                if type(bin_[0]) in CLASSES_DTYPES:
                    bin_type = 'classes'
                else:
                    bin_type = 'edges'

                n_pts = sum(
                    labels[feat].apply(
                        lambda x: is_in_bin(x, bin_, bin_type)
                    )
                )
                if n_pts >= min_pts:
                    good_bins.append(bin_)
                else:
                    small_bins.append(bin_)

            # TODO: More intelligent way to combine small bins
            other_bin = []
            for x in small_bins:
                other_bin.extend(x)

            if len(other_bin) > 0:
                good_bins.append(other_bin)

            seg['bins'] = good_bins
            seg['bin_type'] = bin_type
            seg['n_bins'] = len(good_bins)

        config['segmentation_features'] = seg_features
        self.config = config

    def merge_groups(self, groups):
        '''merge groups on same features'''
        if len(groups) == 1:
            return groups

        # if all groups have different feature-sets return
        feats = [x['features'] for x in groups]
        unique_feats = unique_lists(*feats)
        if len(unique_feats) == len(feats):
            return groups

        new_grps = {}
        for grp in groups:
            idx = [i for i,x in enumerate(unique_feats)
                   if unordered_list_equal(x, grp['features'])][0]
            base = new_grps.get(idx)
            if base is None:
                new_grps[idx] = grp
                continue

            for feat_, bin_, type_ in zip(grp['features'], grp['bins'],
                                          grp['bin_types']):
                idx2 = base['features'].index(feat_)
                base['bins'][idx2].extend(bin_)

        return list(new_grps.values())

    def _load_config_file(self, config_file: str):
        '''
        config must have sections:
            - defaults: min_segment_size (int), features (list[str])
        optional sections:
            - custom_bins:
                key matches features,
                list[list[str|int]] OR list[list[tuple]]
            - n_bins: key matches features, int

        A given feature can EITHER have custom_bins OR n_bins or neither,
        if custom_bins then those are used, if n_bins then n_bins or less are
        created depending on min_segment_size, if neither then bins are
        automatically created
        '''
        config = ConfigParser()
        config.read(config_file)
        feats = eval(config['defaults']['features'])
        config_dict = {
            'min_segment_size': int(config['defaults']['min_segment_size']),
            'features': feats,
        }
        if config.has_section('custom_bins'):
            tmp = {k: eval(v) for k,v in config['custom_bins'].items()}
            config_dict['custom_bins'] = tmp

        if config.has_section('n_bins'):
            tmp = {k: int(v) for k,v in config['n_bins'].items()}
            config_dict['n_bins'] = tmp

        self._load_config_dict(config_dict)

    def _load_config_dict(self, config: dict):
        '''
        dict must have keys:
            - min_segment_size: int
            - features: list[str]
        optional keys are:
            - custom_bins: dict with keys matching feature in features and
              values as list[list[str|int]] or list[list[tuple]]
        '''
        feats = config['features']
        min_pts = config['min_segment_size']
        segments = [{'feature': x} for x in feats]
        if 'custom_bins' in config.keys():
            for key, bins in config['custom_bins'].items():
                if key not in feats:
                    continue

                idx = feats.index(key)
                seg = segments[idx]
                seg['bins'] = bins
                seg['n_bins'] = len(bins)
                if type(bins[0][0]) == tuple:
                    seg['bin_type'] = 'edges'
                else:
                    seg['bin_type'] = 'classes'

        if 'n_bins' in config.keys():
            for key, n_bins in config['n_bins'].items():
                if key not in feats:
                    continue

                idx = feats.index(key)
                if 'bins' in segments[idx].keys():
                    continue

                segments[idx]['n_bins'] = n_bins

        config_dict = {
            'segmentation_features': segments,
            'min_segment_size': min_pts
        }
        self.config = config_dict


def is_in_bin(x, bin_, type_):
    '''is value in bin'''
    if type_=='classes':
        return (x in bin_)
    elif type_=='edges':
        for rng in bin_:
            if rng[0]<=x<=rng[1]:
                return True

        return False
    else:
        raise ValueError('invalid bin type')


def is_in_segment(row, segment: dict):
    '''is data row in segment'''
    feats = segment['features']
    bins = segment['bins']
    types = segment['bin_types']
    if all([is_in_bin(row[feat_], bin_, type_)
            for feat_, bin_, type_ in zip(feats, bins, types)]):
        return True

    return False


def get_bins(dat: pd.Series, n_bins=None):
    # If data is categorical or boolean each category is a bin
    # If data is a number split into even bins
    if dat.dtype in CLASSES_DTYPES:
        groups = list(dat.unique())
        bins = [[x] for x in groups]
        return bins

    if n_bins is not None:
        _, edges = np.histogram(dat, bins=n_bins)
    else:
        _, edges = np.histogram(dat)

    bins = [[(edges[i].item(), edges[i+1].item())]
            for i in range(len(edges)-2)]
    return bins


def unique_lists(*args):
    unique = []
    for a in args:
        if len(unique) == 0:
            unique.append(a)
            continue

        to_add = True
        for b in unique:
            if unordered_list_equal(a,b):
                to_add = False

        if to_add:
            unique.append(a)

    return unique


def unordered_list_equal(a,b):
    return (all([x in a for x in b]) and all([x in b for x in a]))




from typing import Dict

from .model import Model


class PredictiveModel(Model):
    """
    Base class for all predictive models.
    """
    def predict(self, input_data: Dict) -> Dict:
        """
        Implement logic to generate prediction from data

        :params input_data is input data for prediction
        :returns: a dictionary with key `predictions` containing the predictions
        """
        return {'predictions': None}

    def process(self, input_data: Dict) -> Dict:
        """
        Implement logic to generate prediction from data
        """
        return self.predict(input_data=input_data)



import os
import importlib
import pathlib
import inspect
import re
import sys
import logging

from concurrent.futures import ProcessPoolExecutor
from h1st.core.context import discover_h1st_project
from h1st.model.model import Model

logger = logging.getLogger(__name__)
MODEL_FILE_NAME = re.compile(r"^[^_].*_(classifier|detector|model).py$")


def _discover_module(data):
    module_name, filename = data
    result = {}
    try:
        module = importlib.import_module(module_name)
        for attr in dir(module):
            value = getattr(module, attr)

            if inspect.isclass(value) and issubclass(value, Model):
                fullname = value.__name__

                if value.__module__:
                    fullname = f"{value.__module__}.{fullname}"

                sign = (inspect.signature(value))
                params = []
                for v in sign.parameters.values():
                    if v.annotation == str:
                        type_ = 'str'
                    elif v.annotation == int:
                        type_ = 'int'
                    elif v.annotation == float:
                        type_ = 'float'
                    else:
                        type_ = None

                    param = {
                        'name': v.name,
                        'default': v.default if v.default != inspect.Parameter.empty else None,
                        # to be filled by UI later
                        'type': type_,
                        'min': None,
                        'max': None,
                        'choice': [],
                    }

                    params.append(param)

                result[fullname] = {
                    'id': fullname,
                    'name': value.__name__,
                    'tunable': len(params) > 0,
                    'last_modified': os.path.getmtime(filename),
                    'filename': filename,
                    'hyperparameters': params,
                    # to be fillled by UI
                    'selected_metric': None,
                    'options': {},
                }
    except ImportError:
        logger.exception(f'Unable to import module {module_name}')
    except SyntaxError:
        logger.exception(f'Unable to import module {module_name}')

    return result


class ModelExplorer:
    """
    Utility class to discover model class in a project
    """

    def __init__(self, cwd=None):
        self.cwd = cwd

    def discover_models(self):
        project_root, _ = discover_h1st_project(self.cwd)
        if not project_root:
            project_root = os.getcwd()

        if project_root not in sys.path:
            sys.path.append(str(project_root))

        project_root = pathlib.Path(project_root)
        model_files = []

        if (project_root / "__init__.py").exists():
            sys.path.append(str(project_root.parent))
            try:
                importlib.import_module(project_root.name)
                package_root = project_root.name
            except ImportError:
                package_root = ""
        else:
            package_root = ""

        for f in project_root.glob("*.py"):
            if MODEL_FILE_NAME.match(f.name):
                model_files.append(f)

        if (project_root / "models").exists():
            model_files += list((project_root / "models").glob("*.py"))

        # convert to module
        for i, model_file in enumerate(model_files):
            model_package = str(model_file)[len(str(project_root)) + 1:-3]
            model_package = model_package.replace(os.sep, '.')

            if package_root:
                model_package = f"{package_root}.{model_package}"

            model_files[i] = (model_package, model_file)

        result = {}
        with ProcessPoolExecutor(max_workers=4) as pool:
            for res in pool.map(_discover_module, model_files):
                result.update(res)

        return result


if __name__ == "__main__":
    explorer = ModelExplorer()
    print(explorer.discover_models())



from typing import Dict, List
from h1st.model.oracle.oracle_modeler import OracleModeler
from .ts_oracle import TimeSeriesOracle
from h1st.model.oracle.student import AdaBoostModeler, AdaBoostModel, RandomForestModeler, RandomForestModel
from h1st.model.predictive_model import PredictiveModel

class TimeseriesOracleModeler(OracleModeler):
    def __init__(self, teacher: PredictiveModel,
                 ensembler_modeler: PredictiveModel,
                 student_modelers: List = [RandomForestModeler(RandomForestModel),
                                           AdaBoostModeler(AdaBoostModel)],
                 model_class = TimeSeriesOracle
                 ):
        self.teacher = teacher
        self.student_modelers = student_modelers
        self.ensembler_modeler = ensembler_modeler
        self.model_class = model_class
        self.stats = {}

    def build_model(self, data: Dict, id_col: str = None, ts_col: str = None, features: List = None) -> TimeSeriesOracle:
        '''
        Build the oracle
        :param data: a dictionary {'X': ...}
        :param id_col: id column that contains entity ids such as `equipment_id`
        :param ts_col: time-granularity column to group the data
        '''

        self.stats['id_col'] = id_col
        self.stats['ts_col'] = ts_col

        return super().build_model(data, features)


from typing import Dict, List, NoReturn
from sklearn.metrics import confusion_matrix
from sklearn.metrics import precision_recall_fscore_support
from sklearn.metrics import accuracy_score
from sklearn.multioutput import MultiOutputClassifier

from h1st.model.ensemble.stack_ensemble import StackEnsemble
from h1st.model.model import Model


class ClassifierStackEnsemble(StackEnsemble):
    """StackEnsemble classifier

    This is the base class for stack ensemble classifiers
    """

    def __init__(self, ensembler, sub_models: List[Model], **kwargs):
        super().__init__(ensembler, sub_models, **kwargs)


import logging
from typing import Dict, NoReturn

import skfuzzy as fuzz
from skfuzzy import control as ctrl
from skfuzzy.control.term import Term

from h1st.model.rule_based_model import RuleBasedModel

logging.basicConfig(level=logging.INFO)
logger = logging.getLogger(__name__)


class FuzzyLogicModel(RuleBasedModel):
    """
    This FuzzyLogicModel is one kind of rule-based model. You can encode your
    rules in a fuzzy way using variables with membership functions.
    Add variables and use those variables when defining rules.


    """

    def __init__(self):
        self.consequents = []
        self.rules = {}
        self.variables = {}
        self.membership = {
            'triangle': fuzz.trimf,
            'trapezoid': fuzz.trapmf,
            'gaussian': fuzz.gaussmf,
        }
        self.add_variables()
        self.add_rules()
        self.build()

    def add_variables(self) -> NoReturn:
        """
        Add all of your variables with add_variable function.

        .. code-block:: python
            :caption: example

            self.add_variable(
                range_=np.arange(0, 10, 0.5),
                name='sensor1',
                membership_funcs=[('normal', 'gaussian', [3, 3.3]),
                        ('abnormal', 'triangle', [8, 15, 15])],
                type_='antecedent'
            )
            self.add_variable(
                range_=np.arange(0, 10, 0.5),
                name='problem1',
                membership_funcs=[('no', 'trapezoid', [0, 0, 4, 6]),
                        ('yes', 'trapezoid', [4, 6, 10, 10])],
                type_='consequent'
            )
        """

    def add_rules(self) -> NoReturn:
        """
        Add all of your rules with the add_rule function.

        .. code-block:: python
            :caption: example

            self.add_rule(
                'rule1',
                if_=self.variables['sensor1']['abnormal'],
                then_=self.variables['problem1']['yes'])
            self.add_rule(
                'rule2',
                if_=self.variables['sensor1']['normal'],
                then_=self.variables['problem1']['no'])
        """

    def build(self) -> NoReturn:
        ctrl_system = ctrl.ControlSystem(list(self.rules.values()))
        self.css = ctrl.ControlSystemSimulation(ctrl_system)

    def add_rule(self, name, if_: Term, then_: Term) -> NoReturn:
        """
        Add a fuzzy rule. Place antecedent type variables in 'if' statement
        and place consequent type varibles in 'then' statement.

        .. code-block:: python
            :caption: example

            self.add_rule(
                'rule1',
                if_=self.variables['sensor1']['abnormal'],
                then_=self.variables['problem1']['yes'])
        """
        rule = ctrl.Rule(if_, then_)
        self.rules[name] = rule

    def remove_rule(self, name: str) -> NoReturn:
        if name in self.rules:
            del self.rules[name]

    def add_variable(self, range_: list, name: str, membership_funcs: list, type_: str) -> NoReturn:
        """
        Add a variable with their type and membership functions.

        :param range_: the range of variable
        :param name: the name of variable
        :param membership_funcs: this is the list of tuple(membership_func_name, membership_func_type, membership_func_range)
            There are three different kinds of membership_func_type that are supported and their membership_func_range should follow the following formats.
            gaussian: [mean, sigma]
            triangle: [a, b, c] where a <= b <= c
            trapezoid: [a, b, c, d] where a <= b <= c <= d
        :param type_: the varilable type should be either consequent or antecedent

        .. code-block:: python
            :caption: example

            self.add_variable(
                range_=np.arange(0, 10, 0.5),
                name='service_quality',
                membership_funcs=[('Bad', 'gaussian', [2, 1]),
                                  ('Decent', 'triangle', [3, 5, 7]),
                                  ('Great', 'trapezoid', [6, 8, 10, 10])],
                type_='antecedent'
            )
        """

        if type_ == 'antecedent':
            variable = ctrl.Antecedent(range_, name)
        elif type_ == 'consequent':
            variable = ctrl.Consequent(range_, name)
            if name not in self.consequents:
                self.consequents.append(name)
        else:
            logger.warning(f'{type_} is not supported type')
            return None

        for mem_func_name, mem_func_type, mem_func_range in membership_funcs:
            if mem_func_type != 'gaussian':
                variable[mem_func_name] = self.membership[mem_func_type](
                    variable.universe, mem_func_range)
            else:
                variable[mem_func_name] = self.membership[mem_func_type](
                    variable.universe, mem_func_range[0], mem_func_range[1])

        self.variables[name] = variable

    def remove_variable(self, name: str) -> NoReturn:
        if name in self.variables:
            del self.variables[name]

    def visualize(self, variable_name: str) -> NoReturn:
        if variable_name in self.variables:
            self.variables[variable_name].view()
        else:
            logger.warning(f"{variable_name} not found")

    def predict(self, input_data: Dict) -> Dict:
        for key, value in input_data.items():
            self.css.input[key] = value

        self.css.compute()

        outputs = {}
        for cls in self.consequents:
            outputs[cls] = round(self.css.output[cls], 5)

        return outputs



from enum import Enum


class Constituency(Enum):
    """
    A Constituency is the party that is asking for an explanation. This is important to recognize,
    because the same question from different constituencies may warrant different responses: a Data Scientist
    may wish to get a response about detailed model metrics, whereas a Consumer may wish to get a response
    in plain English why her credit application was denied.
    """
    DATA_SCIENTIST = 10
    BUSINESS_MANAGER = 20
    CUSTOMER = 30
    USER = 40
    CONSUMER = 50
    REGULATOR = 60
    LEGISLATOR = 70
    ANY = 99
    OTHER = 100


class Aspect(Enum):
    """
    An Aspect is the type of the question being asked. Different types of questions may warrant very different responses.
    For example, `how do you prevent an autocyber attack in the future?` is an `Operational` type question
    asking you to describe how the system will be used in the event of a future attack.
    Or `what if your model is wrong? - this is a type of `Accountability` question.
    """
    ACCOUNTABLE = 10
    FUNCTIONAL = 20
    OPERATIONAL = 30
    ANY = 99
    OTHER = 100



from h1st.model.predictive_model import PredictiveModel


class Teacher(PredictiveModel):
    """
    Teacher Model in Oracle framework
    """
    pass


class TeacherModeler():
    """
    Knowledge Generalization Modeler
    """
    pass



class Auditable:
    """
    A *Trustworthy-AI* interface that defines the capabilities of objects (e.g., `Models`, `Graphs`)
    that can self-record and report an audit trail of their activities and input/output artifacts.
    For example, an `Auditable` `Model` should be able to record and even retain its training data set,
    the training data scientist, the timestamp, model metrics, etc. available for query at any time in
    the future.
    """

    @property
    def audit_trailing(self):
        return getattr(self, "__audit_trailing", False)

    @audit_trailing.setter
    def audit_trailing(self, value):
        setattr(self, "__audit_trailing", value)

    def audit_trail_get(self, key):
        pass  # to be implemented

    def audit_trail_set(self, key, value):
        if not self.audit_trailing: return
        # to be implemented

    def audit_trail_delete(self, key):
        pass  # to be implemented



from datetime import datetime
import logging
from typing import Dict, List
import ulid

import numpy as np
import pandas as pd

from h1st.model.predictive_model import PredictiveModel
from h1st.model.model import Model


class KSWE(PredictiveModel):
    def __init__(self):
        super().__init__()
        self.stats = {}
        self.segmentor = None
        self.sub_models = None
        self.ensemble = None

    def predict(self, input_data: Dict):

        # Segment data
        if isinstance(input_data['X'], pd.DataFrame):
            input_data['X'].reset_index(inplace=True, drop=True)
        segmented_data = self.segmentor.process(input_data)['segment_data']

        # Generate sub_model predictions
        sub_model_predictions = {}
        for name, data in segmented_data.items():
            if name not in self.sub_models:
                logging.info('There is no model for this data segment.'
                             'Generate the output as NaN for this data segment.') 
                nan_output = np.empty(data.shape[0]) 
                nan_output[:] = np.nan
                predictions = nan_output
            else:
                # generate predictions from sub_models
                predictions = self.sub_models[name].predict(
                    {'X': data})['predictions']

            sub_model_predictions[name] = {
                'index': data.index, 
                'predictions': predictions,
                'num_of_data':input_data['X'].shape[0],
            }

        # Generage ensemble's prediction
        final_predictions = self.ensemble.predict(sub_model_predictions)

        return final_predictions

    def persist(self, version: str=None) -> None:
        if version is None:
            now = datetime.utcnow()
            now_str = now.strftime('%Y%m%d-%H%M')
            version = f'{str(ulid.new())[:4]}-{now_str}'

        self.stats['version'] = version
        self.stats['ensemble_version'] = self.ensemble.persist(version+'_ensemble')
        self.stats['ensemble_class'] = self.ensemble.__class__
        self.stats['segmentor_version'] = self.segmentor.persist(version+'_segmentor')
        self.stats['segmentor_class'] = self.segmentor.__class__

        sub_model_info = {}
        for name, sub_model in self.sub_models.items():
            v = sub_model.persist(version+f'_{name}')
            sub_model_info[name] = {'version': v, 'model_class': sub_model.__class__}
        self.stats['sub_model_info'] = sub_model_info
        super().persist(version)
        return version

    def load_params(self, version: str=None) -> None:
        super().load_params(version)
        ensemble_class = self.stats['ensemble_class']
        ensemble_version = self.stats['ensemble_version']
        self.ensemble = ensemble_class().load_params(ensemble_version)

        segmentor_class = self.stats['segmentor_class']
        segmentor_version = self.stats['segmentor_version']
        self.segmentor = segmentor_class().load_params(segmentor_version)

        sub_models = {}
        for name, info in self.stats['sub_model_info'].items():
            sub_model = info['model_class']().load_params(info['version'])
            sub_models[name] = sub_model
        self.sub_models = sub_models
        return self

    @classmethod
    def construct_model(cls, segmentor: Model, sub_models: dict, ensemble: Model):
        model = cls()
        model.segmentor = segmentor
        model.sub_models = sub_models
        model.ensemble = ensemble
        return model


    # def persist(self, version: str) -> None:
    #     self.ensemble.persist(version+'_ensemble')
    #     sub_model_names = []
    #     for name, sub_model in self.sub_models.items():
    #         sub_model.persist(version+f'_{name}')
    #         sub_model_names.append(name)
    #     self.stats['sub_model_names'] = sub_model_names
    #     super().persist(version)

    # def load_params(self, version: str) -> None:
    #     super().load_params(version)
    #     sub_model_names = self.stats['sub_model_names']
    #     self.ensemble.load_params(version+'_ensemble')
    #     sub_models = {}
    #     for name in sub_model_names:
    #         sub_model = self.sub_model_class().load_params(version+f'_{name}')
    #         sub_models[name] = sub_model
    #     self.sub_models = sub_models





from h1st.model.predictive_model import PredictiveModel
from h1st.model.rule_based_model import RuleBasedModel


class BooleanModel(PredictiveModel):
    """
    In H1st AI, we intend for BooleanModel to mean predictive boolean model. It is somewhat pointless to have a
    "non-predictive" boolean model, since that is basically "just code".
    """
    pass



from .explainable import Explainable
from .describable import Describable
from .auditable import Auditable
from .debiasable import Debiasable


# __TODO__: implement AI-RULE interfaces


class Trustable(Auditable, Debiasable, Describable, Explainable):
    """
    Base class for all `Trustable` interfaces
    """



class HasWebUI():
    def handle_request(self, req):
        if (req.method == 'GET'):
            return self.handle_get(req)
        elif (req.method == 'POST'):
            return self.handle_post(req)
        elif (req.method == 'PUT'):
            return self.handle_post(req)
        elif (req.method == 'DELETE'):
            return self.handle_post(req)
        else:
            return self.handle_default(req)

    def handle_get(self, req):
        raise NotImplementedError('Please implement get method handler')

    def handle_post(self, req):
        raise NotImplementedError('Please implement post method handler')

    def handle_put(self, req):
        raise NotImplementedError('Please implement put method handler')

    def handle_delete(self, req):
        raise NotImplementedError('Please implement delete method handler')

    def handle_default(self, req):
        raise NotImplementedError('Method is not supported')



from typing import Dict
import pandas as pd
from h1st.model.predictive_model import PredictiveModel


class Ensemble(PredictiveModel):
    """
    Ensemble Model in Oracle framework
    """

    def combine(self, teacher_pred, student_preds):
        '''
        Combine output of teacher and students using majority vote by default. In case
        when majority vote cannot be applied, use teacher's output as the final output.
        Inherit and override this method to use your custom combining approach.
        :param teacher_pred: teacher's prediction
        :param student_pred: student's prediction
        :returns: a dictionary with key `predictions` containing the predictions
        '''
        if len(student_preds) == 1:
            predictions = student_preds[0]
        else:
            pred_dict = {'teacher_pred': teacher_pred}
            pred_dict.update({f'student_pred_{i}': student_pred
                              for i, student_pred in enumerate(student_preds)})
            predictions = pd.DataFrame(pred_dict).mode(axis='columns', numeric_only=True)[0]

        return {'predictions': predictions}

    def predict(self, input_data: Dict) -> Dict:
        teacher_pred = pd.Series(input_data['teacher_pred'])
        student_preds = [pd.Series(student_pred) for student_pred in input_data['student_preds']]
        return self.combine(teacher_pred, student_preds)



from typing import Any, Dict, List
import pandas as pd
from h1st.model.ml_modeler import MLModeler

from h1st.model.modeler import Modeler
from h1st.model.oracle.oracle import Oracle
from h1st.model.oracle.student import AdaBoostModeler, RandomForestModeler
from h1st.model.predictive_model import PredictiveModel
from h1st.model.ensemble.stack_ensemble_modeler import StackEnsembleModeler

class OracleModeler(Modeler):
    def __init__(self, teacher: PredictiveModel,
                 ensembler_modeler: Modeler,
                 student_modelers: List = [RandomForestModeler(), AdaBoostModeler()],
                 model_class = Oracle
                 ):
        self.teacher = teacher
        self.student_modelers = student_modelers
        self.ensembler_modeler = ensembler_modeler
        self.model_class = model_class
        self.stats = {}

    def build_model(self, data: Dict[str, Any] = None, features: List = None) -> Oracle:
        """
        Build the student and ensemble components.
        :param data: Unlabeled data.
        """
        self.stats['features'] = features
        # Generate features to get students' predictions
        train_data = self.model_class.generate_data({'X': data['unlabeled_data']}, self.teacher, self.stats)

        # Train the student model
        self.students = [student_modeler.build_model(train_data)
                         for student_modeler in self.student_modelers]

        # Train the ensembler
        labeled_data = data.get('labeled_data', None)

        if isinstance(self.ensembler_modeler, MLModeler) and labeled_data is None:
            raise ValueError('No data to train the machine-learning-based ensemble')

        if labeled_data is not None:
            ensembler_train_data = self.model_class.generate_data({'X': labeled_data['X_train']}, self.teacher, self.stats)
            ensembler_test_data = self.model_class.generate_data({'X': labeled_data['X_test']}, self.teacher, self.stats)
            student_preds_train_data = [pd.Series(student.predict(ensembler_train_data)['predictions']) for student in self.students]
            student_preds_test_data = [pd.Series(student.predict(ensembler_test_data)['predictions']) for student in self.students]
            ensembler_data = {'X_train': pd.concat(student_preds_train_data + 
                                                [ensembler_train_data['y']], axis=1),
                                    'y_train': labeled_data['y_train'],
                                    'X_test': pd.concat(student_preds_test_data + 
                                                [ensembler_test_data['y']], axis=1),
                                    'y_test': labeled_data['y_test'],
                                    }
        else:
            ensembler_data = None

        ensembler = self.ensembler_modeler.build_model(ensembler_data)
        oracle = self.model_class(self.teacher, self.students, ensembler)
        # Pass stats to the model
        if self.stats is not None:
            oracle.stats = self.stats.copy()

        return oracle



"""
Oracle architecture:

@startuml
allowmixing

Component Oracle #EEE {
	Class Teacher
	Class Student
	Class Ensemble
}

Actor "AI Engineer" as User

User .down.> TeacherModeler : uses
User .down.> StudentModeler : uses

TeacherModeler -down-> Teacher : builds
StudentModeler -down-> Student : builds
Teacher .right.> Student : teaches



Teacher -down-> Ensemble : trains
Student -down-> Ensemble : trains

Note as N1  #green
<size:16><color:white>Construction Phase</color></size>
end Note
@enduml

@startuml
allowmixing

Component Oracle #EEE {
    Class Teacher <<RuleBasedModel>>
    Class Student <<ML Generalizer>>
	Class Ensemble
}

Database Data
Circle Prediction

Data -down-> Teacher
Data -down-> Student

Teacher -down-> Ensemble : votes
Student -down-> Ensemble : votes

Ensemble -down-> Prediction

Note as N1  #green
<size:16><color:white>Execution Phase</color></size>
end Note
@enduml
"""

from typing import Dict, List
import pandas as pd
from h1st.model.ensemble.stack_ensemble import StackEnsemble
from h1st.model.predictive_model import PredictiveModel


class Oracle(PredictiveModel):
    """
    Oracle Model in Oracle framework
    """

    def __init__(self, 
                 teacher: PredictiveModel,
                 students: PredictiveModel,
                 ensembler: PredictiveModel):
        """
        :param teacher: The knowledge model.
        :param student_modelers: The student modelers.
        :param ensemble: The ensemble model class.
        """
        self.teacher = teacher
        self.students = students
        self.ensembler = ensembler
        self.stats = {}

    @classmethod
    def generate_features(cls, data: Dict):
        return {'data': data['data'].copy()}

    @classmethod
    def generate_data(cls, data: Dict, teacher: PredictiveModel, stats: Dict) -> Dict:
        """
        Generate data to train Student models.
        Return a copy of the provided data by default.
        Override this function to implement custom data generation
        :param data: unlabelled data.
        :returns: a dictionary of features and teacher's prediction.
        """
        if 'X' not in data:
            raise ValueError('Please provide data in form of {\'X\': pd.DataFrame}')

        df = data['X']

        features = stats['features']
        if features is not None:
            df = df[features]

        df = cls.generate_features({'data': df})['data']

        teacher_pred = teacher.predict({'X': df})
        if 'predictions' not in teacher_pred:
            raise KeyError('Teacher\'s output must contain a key named `predictions`')
        return {'X': df.copy(), 'y': pd.Series(teacher_pred['predictions'])}

    def predict(self, input_data: Dict) -> Dict:
        """
        Implement logic to generate prediction from data. The Oracle expects the same features provided during `build` phase to be in the provided data. It automatically process the data the same way to that of the `build` phase.

        :params input_data: an dictionary with key `X` containing the data to get predictions.
        :returns: a dictionary with key `predictions` containing the predictions
        """
        if not hasattr(self, 'students'):
            raise RuntimeError('No student built')

        # Generate features to get students' predictions
        predict_data = self.__class__.generate_data(input_data, self.teacher, self.stats)

        # Generate student models' predictions
        student_preds = [pd.Series(student.predict(predict_data)['predictions']) for student in self.students]

        return self.ensembler.predict({'X': pd.concat(student_preds + [predict_data['y']], axis=1)})

    def persist(self, version=None):
        version = self.ensembler.persist(version)

        for student in self.students:
            version = student.persist(version)
        super().persist(version)

    def load_params(self, version: str = None) -> None:
        self.ensembler.load_params(version)
        for student in self.students:
            student.load_params(self.ensembler.version)
        super().load_params(version)



from typing import Any

from .predictive_model import PredictiveModel


class MLModel(PredictiveModel):
    """
    Base class for H1st Model.

    To create your own Machine Learning model, inherit `MLModel` class and implement `prepare_data`, 
    `train` and `predict` accordingly. Please refer to Tutorial for more details how to create a model.

    The framework allows you to persist and load model to the model repository.
    To persist the model, you can call `persist()`, and then `load_params()` to retrieve the model.
    See `persist()` and `load_params()` document for more detail.

        .. code-block:: python
            :caption: Model Persistence and Loading Example

            import os
            import tempfile
            from typing import Any, Dict

            from h1st.model.ml_model import MLModel
            from h1st.model.ml_modeler import MLModeler
            from sklearn.datasets import load_iris
            from sklearn.model_selection import train_test_split
            from sklearn.ensemble import RandomForestClassifie

            class MyMLModel(MLModel):
                def predict(self, input_data: Dict) -> Dict:
            return {'predictions': self.base_model.predict(input_data['X'])}

            class MyMLModeler(MLModeler):
                def __init__(self):
                    self.model_class = MyMLModel

                def train_base_model(self, prepared_data: Dict[str, Any]) -> Any:
                    X, y = prepared_data['X'], prepared_data['y']
                    model = RandomForestClassifier(random_state=0)
                    model.fit(X, y)
                    return model

            X, y = load_iris(return_X_y=True)
            X_train, X_test, y_train, y_test = train_test_split(X, y, test_size=0.33, random_state=42)
            prepared_data = {'X': X_train, 'y': y_train}
            my_modeler = MyMLModeler()
            my_model = my_modeler.build_model(prepared_data)

            y_pred = my_model.predict({'X': X_test})['predictions']
            accuracy = accuracy_score(y_test, y_pred)
            print("Accuracy (test): %0.1f%% " % (accuracy * 100))

            with tempfile.TemporaryDirectory() as path:
                # Set the model repository to the temporary directory
                os.environ['H1ST_MODEL_REPO_PATH'] = path
                my_model.persist('1st_version')

                # Load the model from the repo
                my_model_2 = MyMLModel()
                my_model_2.load_params('1st_version')
                y_pred = my_model_2.predict({'X': X_test})['predictions']
                accuracy = accuracy_score(y_test, y_pred)
                print("Accuracy (test): %0.1f%% " % (accuracy * 100))
    """

    @property
    def base_model(self) -> Any:
        return getattr(self, "__base_model", None)

    @base_model.setter
    def base_model(self, value):
        setattr(self, "__base_model", value)



import shap


class SHAPModelDescriber:
    def __init__(self, model, data):
        self.model = model
        self.data = data
        self.get_shap_describer()
        self.get_shap_values()
        ## Only for testing if plots are generated
        # self.plot_shap_describer()

    def plot_shap_describer(self):
        shap.summary_plot(self.shap_values, self.data['train_df'])

    def get_shap_describer(self):
        self.shap_describer = shap.TreeExplainer(self.model)

    def get_shap_values(self):
        self.shap_values = self.shap_describer.shap_values(
            self.data["train_df"], check_additivity=False
        )



from typing import Dict
import pandas as pd
from .predictive_model import PredictiveModel


class RuleBasedModel(PredictiveModel):
    pass

class RuleBasedRegressionModel(RuleBasedModel):
    """
    Combine predictions using averaging strategy by default.
    """
    def predict(self, input_data: Dict) -> Dict:
        """
        :param input_data: a dictionary of predictions.
        :returns: a pandas series of averaged values per group of predictions.
        """
        predictions = pd.DataFrame(input_data['X']).mean(axis=1,
                                                    skipna=True,
                                                    numeric_only=True).values
        return {'predictions': predictions}

class RuleBasedClassificationModel(RuleBasedModel):
    """
    Combine predictions using majority voting by default.
    """
    def predict(self, input_data: Dict) -> Dict:
        """
        :param input_data: a dictionary of predictions.
        :returns: a pandas series of the most popular values per group of predictions.
        """
        predictions = pd.DataFrame(input_data['X']).mode(axis='columns',
                                                    numeric_only=True)[0].values
        return {'predictions': predictions}



from .lime_model_explainer import LIMEModelExplainer
from .enums import Constituency, Aspect
from .explainer import Explainer


class Explainable:
    """
    A *Trustworthy-AI* interface that defines the capabilities of objects (e.g., `Models`, `Graphs`)
    that are `Explainable`, i.e., they can self-explain certain decisions they have made. For example,
    an `Explainable` `Model` should be able to provide a detailed explanation of a specific decision
    it made at some specified time or on a given set of inputs in the past.
    """

    def explain(self, decision=None, constituent=Constituency.ANY, aspect=Aspect.ANY):
        """
        Returns an explanation for a decision made by the Model based on `Who's asking` and `why`.

            Parameters:
                constituent : Constituency: The `Constituency` asking for the explanation
                aspect : Aspect: The `Aspect` of the question (e.g., Accountable, Functional, Operational)
                decision : array-like: the input data of the decision to be explained

            Returns:
                out : Specific decision explanation (e.g., SHAP or LIME)
        """
        explainer = Explainer(self, decision)
        explainer.lime_explainer = LIMEModelExplainer(
            decision, self._native_model, self.prepared_data
        )
        explainer.generate_report(decision, constituent, aspect)
        return explainer



from typing import Any, Dict
from copy import deepcopy

import pandas as pd
from sklearn.ensemble import RandomForestClassifier as SKRandomForestClassifier
from sklearn import metrics

from h1st.model.ml_model import MLModel
from h1st.model.ml_modeler import MLModeler


class RandomForestClassifier(MLModel):
    def predict(self, input_data: Dict) -> Dict:
        y = self.base_model.predict(input_data['X'])
        return {'predictions': y}


class RandomForestClassifierModeler(MLModeler):
    def __init__(self, model_class=RandomForestClassifier):
        self.model_class = model_class
        self.stats = {}
    
    def evaluate_model(self, data: Dict, model: MLModel) -> Dict:
        super().evaluate_model(data, model)
        X, y_true = data['X_test'], data['y_test']
        y_pred = pd.Series(model.predict({'X': X, 'y': y_true})['predictions'])
        return {'r2_score': metrics.r2_score(y_true, y_pred)}

    def train_base_model(self, data: Dict[str, Any]) -> Any:
        X, y = data['X_train'], data['y_train']
        model = SKRandomForestClassifier()
        model.fit(X, y)
        return model




from typing import Dict, List, Any

import numpy as np
from sklearn.preprocessing import RobustScaler
from h1st.model.ensemble.stack_ensemble import StackEnsemble
from h1st.model.ml_modeler import MLModeler
from h1st.model.predictive_model import PredictiveModel
from h1st.model.ensemble.ensemble_modeler import EnsembleModeler
from h1st.model.model import Model


class StackEnsembleModeler(EnsembleModeler):
    """
    Base StackEnsemble class to implement StackEnsemble classifiers or regressors.
    """

    def __init__(self, ensembler_modeler, sub_models: List[PredictiveModel], **kwargs):
        """
        :param **kwargs: StackEnsemble use h1st models as submodels and the predict function of
        h1st model receives a dictionary and returns a dictionary. users can set the key of these
        dictionaries with the following keyword arguments.

            **submodel_predict_input_key (str): the default value of submodel_input_key is 'X' \n
            **submodel_predict_output_key (str): the default value of submodel_output_key is 'predictions'
        """
        self.ensembler_modeler = ensembler_modeler
        self._sub_models = sub_models
        self._submodel_predict_input_key = kwargs.get('submodel_predict_input_key', 'X')
        self._submodel_predict_output_key = kwargs.get('submodel_predict_output_key', 'predictions')
        self.model_class = StackEnsemble
        

    def build_model(self, data: Dict[str, Any] = None) -> StackEnsemble:
        """
        Trains an ensembler with input is merging raw data 'X_train' and predictions of sub models, with targets 'y_train'
        'X_train' is supposed to be numeric.
        :param prepared_data: output of prep() implemented in subclass of StackEnsemble, its structure must be this dictionary
            {
                'X_train': ...,
                'X_test': ...,
                'y_train': ...,
                'y_test': ...
            }
        """
        if data is None:
            data = self.load_data()

        if data is not None:
            X_train, y_train = (data['X_train'], data['y_train'])
            X_train = self.model_class.preprocess(self._sub_models,
                                                self._submodel_predict_input_key,
                                                self._submodel_predict_output_key,
                                                X_train
                                                )

            X_test, y_test = (data['X_test'], data['y_test'])
            X_test = self.model_class.preprocess(self._sub_models,
                                                self._submodel_predict_input_key,
                                                self._submodel_predict_output_key,
                                                X_test
                                                )

            ensembler = self.ensembler_modeler.build_model({'X_train': X_train, 'y_train': y_train,
                                                            'X_test': X_test, 'y_test': y_test,
                                                            })
        else:
            ensembler = self.ensembler_modeler.build_model()
        ensemble = self.model_class(ensembler=ensembler,
                                    sub_models=self._sub_models,
                                    submodel_predict_input_key=self._submodel_predict_input_key,
                                    submodel_predict_output_key=self._submodel_predict_output_key)
        return ensemble



from typing import Any, Dict
from copy import deepcopy
from sklearn.linear_model import LogisticRegression
from sklearn.ensemble import RandomForestClassifier, AdaBoostClassifier
from h1st.model.ml_model import MLModel
from h1st.model.ml_modeler import MLModeler


class Student(MLModel):
    """
    Knowledge Generalization Model
    """

    def predict(self, input_data: Dict) -> Dict:
        """
        Implement logic to generate prediction from data

        :params input_data: an dictionary with key `X` containing the data to get predictions.
        :returns: a dictionary with key `predictions` containing the predictions
        """
        return {'predictions': self.base_model.predict(input_data['X'])}


class StudentModeler(MLModeler):
    """
    Knowledge Generalization Modeler
    """

    def __init__(self, base_model=LogisticRegression()):
        self.model_class = Student
        self.base_model = base_model

    def train_base_model(self, prepared_data: Dict[str, Any]) -> Any:
        self.base_model.fit(prepared_data['X'], prepared_data['y'])
        return deepcopy(self.base_model)


class RandomForestModel(Student):
    """
    Knowledge Generalization Model backed by a RandomForest algorithm
    """
    pass


class RandomForestModeler(StudentModeler):
    """
    Knowledge Generalization Modeler backed by a RandomForest algorithm.
    """

    def __init__(self, base_model=RandomForestClassifier()):
        super().__init__(base_model=base_model)
        self.model_class = RandomForestModel


class AdaBoostModel(Student):
    """
    Knowledge Generalization Model backed by an AdaBoost algorithm
    """
    pass


class AdaBoostModeler(StudentModeler):
    """
    Knowledge Generalization Modeler backed by a AdaBoost algorithm.
    """

    def __init__(self, base_model=AdaBoostClassifier()):
        super().__init__(base_model=base_model)
        self.model_class = AdaBoostModel



from copy import deepcopy
import json
import logging
from typing import Dict, List

import pandas as pd
from sklearn.model_selection import train_test_split as sk_train_test_split

from h1st.model.modeler import Modeler
from h1st.model.rule_based_modeler import RuleBasedModeler
from .ensemble import MajorityVotingEnsemble
from .kswe import KSWE
from .segmentor import CombinationSegmentor, MaxSegmentationModeler
from .sub_model_modeler import RandomForestClassifierModeler


class KSWEModeler(Modeler):
    def __init__(self, model_class=KSWE):
        self.model_class = model_class
        self.stats = {}

    def train_test_split(self, dataframe, test_size=0.4) -> Dict:
        sk_train_test_split(dataframe, test_size)
        pass 

    def build_model(
        self,
        input_data: Dict, 
        sub_model_modeler: Modeler = RandomForestClassifierModeler(),
        ensemble_modeler: Modeler = RuleBasedModeler(model_class=MajorityVotingEnsemble),
        segmentor_modeler: Modeler = MaxSegmentationModeler(),
        segmentation_config: Dict = None, 
    ):
        '''
        :param input_data: 
            Data for training and evaluating sub_models and ensembles of KSWE.
        :param sub_model_modeler:
            User can use their own sub_model modeler, or use pre-built modeler,
            such as RandomForestClassifierModeler.
        :param ensemble_modeler:
            User can use their own ensemble modeler, or use pre-built modeler,
            such as MajorityVotingEnsembleModeler.
        :param segmentor:
            Segment data into segmented data based on segmentation_configs
        :param segmentation_config:
            Domain knowledge that will be used for segmenting data and build KSWE.
            Either a dict where each key is a column to segment on and the
            value is a list[list], each level 0 list item represents a bin and
            each level 1 list item is a class of range in that bin
            i.e.:
                {'time_of_day': [['day'], ['night']],
                'month': [['Jan', 'Feb'], ['Mar', 'Apr']],
                'depth': [[(0, 50), (75,100)], [(51, 74)]]}
            Alternatively, this can be string path to config file, see example for format.
        '''
        if 'X_train' not in input_data or 'X_test' not in input_data:
            raise KeyError('key "X_train" or "X_test" are not in your input_data')

        if (bool('y_train' in input_data) ^ bool('y_test' in input_data)): 
            raise KeyError('key "y_train" or "y_test" are not in your input_data')

        if not isinstance(input_data['X_train'], pd.DataFrame):
            raise TypeError(f'input_data["X_train"] should be pandas.DataFrame \
                            but {type(input_data["X_train"])} were given')
        
        if 'y_train' in input_data and 'y_test' in input_data:
            if input_data['X_train'].shape[0] != input_data['y_train'].shape[0]:
                raise ValueError(f'The length of X_train {input_data["X_train"].shape[0]}'
                                f'and y_train {input_data["y_train"].shape[0]} are different.')
            if input_data['X_test'].shape[0] != input_data['y_test'].shape[0]:
                raise ValueError(f'The length of X_test {input_data["X_test"].shape[0]}'
                                f'and y_test {input_data["y_test"].shape[0]} are different.')                    

        segmentor = segmentor_modeler.build_model(input_data, segmentation_config)
        segmentor_output = segmentor.process({'X': input_data['X_train']})

        # Train sub_models
        sub_models = {}
        segmented_data = {}
        for name, segmented_X_train in segmentor_output['segment_data'].items():
            train_test_data = self.prepare_sub_model_train_test_data(
                name, segmented_X_train, input_data)
            
            # TODO: Move this validation into segmentor_modeler
            if ('y_train' in train_test_data) and (train_test_data['y_train'].nunique() == 1):
                logging.info(f'Skip the Training of sub_model {name} because '
                             'there is only one y class or one constant value.')
                # remove this data segment from segmentor
                del segmentor.stats['segment_info'][name]
                continue

            segmented_data[name] = train_test_data
            logging.info(f'Training sub_model {name} based on '
                         f'{segmented_X_train.shape[0]} samples')
            sub_model = sub_model_modeler.build_model(train_test_data)
            sub_model.stats['segment_info'] = segmentor_output['segment_info'][name]
            sub_models[name] = sub_model
            logging.info(f'Metrics for trained submodel {name}:\n'
                         f'{json.dumps(sub_model.metrics)}')

        # Save segmentation_logics in self.stats. We will use this in KSWE predict method.
        self.stats['segment_info'] = segmentor_output['segment_info']
        self.stats['segmentation_config'] = segmentation_config

        # Build ensemble
        ensemble = self.build_ensemble(
            ensemble_modeler, sub_models, segmented_data)

        # Build KSWE
        kswe = self.model_class.construct_model(
            segmentor=segmentor, 
            sub_models=sub_models, 
            ensemble=ensemble
        )
        kswe.stats.update(self.stats)
        kswe.metrics = sub_model_modeler.evaluate_model(input_data, kswe)

        # Show the evaluation results of KSWE
        print('kswe test results:', kswe.metrics)
        return kswe

    def build_ensemble(self, ensemble_modeler, sub_models, segmented_data):
        ensemble_training_data = {}
        for name, sub_model in sub_models.items():
            data_segment = segmented_data[name]

            # Generate sub_model predictions
            sub_model_pred = sub_model.predict(
                {'X': data_segment['X_train']})['predictions']
            ensemble_training_data[name] = {
                'X_train': data_segment['X_train'],
                'y_pred_sub_model': sub_model_pred
            }

            # If y_train exists, add it in training data
            if 'y_train' in data_segment:
                ensemble_training_data[name]['y_train'] = data_segment['y_train']

        # Build Ensemble            
        ensemble = ensemble_modeler.build_model(ensemble_training_data)
        return ensemble

    def prepare_sub_model_train_test_data(self, name, X_train, input_data):
        train_test_data = {
            'X_train': X_train,
            'X_test': deepcopy(input_data['X_test']),
        }
        if 'y_test' in input_data:
            train_test_data['y_train'] = input_data['y_train'].loc[X_train.index]
            train_test_data['y_test'] = deepcopy(input_data['y_test'])
        return train_test_data
            



from typing import Dict, List, Any
import numpy as np
from h1st.model.ensemble.ensemble import Ensemble
from h1st.model.predictive_model import PredictiveModel


class StackEnsemble(Ensemble):
    """
    Base StackEnsemble class to implement StackEnsemble classifiers or regressors.
    """
    def __init__(self, ensembler, sub_models: List[PredictiveModel], **kwargs):
        """
        :param **kwargs: StackEnsemble use h1st models as submodels and the predict function of
        h1st model receives a dictionary and returns a dictionary. users can set the key of these
        dictionaries with the following keyword arguments.

            **submodel_predict_input_key (str): the default value of submodel_input_key is 'X' \n
            **submodel_predict_output_key (str): the default value of submodel_output_key is 'predictions'
        """
        self.ensembler = ensembler
        self._sub_models = sub_models
        self._submodel_predict_input_key = kwargs.get('submodel_predict_input_key', 'X')
        self._submodel_predict_output_key = kwargs.get('submodel_predict_output_key', 'predictions')

    @classmethod
    def preprocess(cls, sub_models: List[PredictiveModel],
                        submodel_predict_input_key: str,
                        submodel_predict_output_key: str,
                        X: Any) -> Any:
        """
        Predicts all sub models then merge input raw data with predictions for ensembler's training or prediction
        :param X: raw input data
        """
        if isinstance(X, list):
            predictions = [np.hstack((x, StackEnsemble.get_submodels_prediction(sub_models,
                                                                  submodel_predict_input_key,
                                                                  submodel_predict_output_key, x)))
                                                                for x in X]
            return np.vstack(predictions)
        else:
            return np.hstack((X, StackEnsemble.get_submodels_prediction(sub_models,
                                                        submodel_predict_input_key,
                                                        submodel_predict_output_key,
                                                        X
                                                        )))

    @classmethod
    def get_submodels_prediction(cls, sub_models: List[PredictiveModel], submodel_predict_input_key: str, submodel_predict_output_key: str, X: Any) -> Any:
        preds = []
        for m in sub_models:
            pred = m.predict({submodel_predict_input_key: X})
            output_key = pred.get(submodel_predict_output_key)
            if output_key is not None:
                preds.append(output_key.reshape(-1,1))
        return np.hstack(preds)

    def predict(self, data: Dict) -> Dict:
        """
        Returns predictions of ensembler model for input raw data combining with predictions from sub models

        This method will raise ModelException if ensembler model has not been trained.

        :param data: must be a dictionary {'X': ...}, with X is raw data (numeric only)
        :return:
            output will be a dictionary {'predictions': ....}
        """
        X = StackEnsemble.preprocess(self._sub_models,
                                     self._submodel_predict_input_key,
                                     self._submodel_predict_output_key,
                                     data['X']
                                    )
        return {'predictions': self.ensembler.predict({'X': X})['predictions']}

    def persist(self, version=None):
        version = self.ensembler.persist(version)
        version = super().persist(version)
        return version

    def load_params(self, version: str = None) -> None:
        self.ensembler.load_params(version)
        super().load_params(self.ensembler.version)



import numpy as np
import lime.lime_tabular as lt


class LIMEModelExplainer:
    def __init__(self, decision, model, data):
        self.model = model
        self.data = data
        self.feature_names = list(self.data["train_df"].columns)
        self.decision_input = decision[0]
        self.lime_explainer()
        self.explain_decision()
        ## Only for testing if plots are generated
        # self.plot_lime_explanation()

    def plot_lime_explanation(self):
        self.explanation.show_in_notebook(show_table=True, show_all=False)

    def explain_decision(self):
        self.explanation = self.explainer.explain_instance(
            self.decision_input,
            self.model.predict,
        )

    def lime_explainer(self):
        self.explainer = lt.LimeTabularExplainer(
            np.array(self.data["train_df"]),
            feature_names=self.feature_names,
            verbose=False,
            mode="regression",
        )



from typing import Any, Dict
from .ml_model import MLModel
from .modeler import Modeler


class MLModeler(Modeler):
    """
    Base class for H1st ML Modelers. Has capabilities that are specific to MLModels.
    """

    def train_base_model(self, prepared_data: Dict[str, Any]) -> Any:
        """
        Implement logic of training the base/native model

        :param prepared_data: prepared data
        """

    def train_model(self, data: Dict[str, Any] = None) -> MLModel:
        """
        Implement logic to create the corresponding MLModel, including both training and evaluation.
        """
        if self.model_class is None:
            raise ValueError('Model class not provided')

        if not data:
            data = self.load_data()
        
        base_model = self.train_base_model(data)

        ml_model = self.model_class()
        ml_model.base_model = base_model

        # Pass stats to the model
        if self.stats is not None:
            ml_model.stats = self.stats.copy()
        # Compute metrics and pass to the model
        ml_model.metrics = self.evaluate_model(data, ml_model)
        return ml_model

    def build_model(self, data: Dict[str, Any] = None) -> MLModel:
        return self.train_model(data)


from typing import Any, Dict
from h1st.model.modelable import Modelable
from h1st.model.rule_based_model import RuleBasedModel
from .modeler import Modeler

class RuleBasedModeler(Modeler):
    def __init__(self, model_class=RuleBasedModel):
        self.model_class = model_class

    def build_model(self, data: Dict[str, Any] = None) -> RuleBasedModel:
        return self.model_class()


from typing import Union, Any, NoReturn
from abc import ABC, abstractmethod


class Storage(ABC):
    """
    Base class for storage
    """

    @abstractmethod
    def get_obj(self, name: str) -> Any:
        ...

    @abstractmethod
    def set_obj(self, name: str, value: Any) -> NoReturn:
        ...

    @abstractmethod
    def get_bytes(self, name: str) -> bytes:
        ...

    @abstractmethod
    def set_bytes(self, name: str, value: bytes) -> NoReturn:
        ...

    @abstractmethod
    def exists(self, name: str) -> bool:
        ...

    @abstractmethod
    def delete(self, name: str) -> Any:
        ...

    def delete_namespace(self, namespace: str):
        raise NotImplementedError()

    def list_keys(self, namespace: str) -> list:
        raise NotImplementedError()



# -*- coding: utf-8 -*-
from setuptools import setup

packages = \
['h1st',
 'h1st.core',
 'h1st.exceptions',
 'h1st.h1flow',
 'h1st.h1flow.ui',
 'h1st.model',
 'h1st.model.ensemble',
 'h1st.model.kswe',
 'h1st.model.oracle',
 'h1st.model.repository',
 'h1st.model.repository.storage',
 'h1st.trust']

package_data = \
{'': ['*']}

install_requires = \
['graphviz>=0.19.1,<0.20.0',
 'lime>=0.2.0.1,<0.3.0.0',
 'pyarrow>=7.0.0,<8.0.0',
 'pyyaml>=6.0,<7.0',
 's3fs>=2022.1.0,<2023.0.0',
 'scikit-fuzzy>=0.4.2,<0.5.0',
 'scikit-learn>=1.0.2,<2.0.0',
 'shap>=0.40.0,<0.41.0',
 'ulid-py>=1.1.0,<2.0.0']

setup_kwargs = {
    'name': 'h1st',
    'version': '0.1.9',
    'description': 'Human-First AI (H1ST)',
    'long_description': '## Join the Human-First AI revolution\n_“We humans have .. insight that can then be mixed with powerful AI .. to help move society forward. Second, we also have to build trust directly into our technology .. And third, all of the technology we build must be inclusive and respectful to everyone.”_\n<br/>— Satya Nadella, Microsoft CEO\n\nAs trail-blazers in Industrial AI, our team at Arimo-Panasonic has found Satya Nadella‘s observations to be powerful and prescient. Many hard-won lessons from the field have led us to adopt this approach which we call Human-First AI (`H1st` AI). \n\nToday, we‘re excited to share these ideas and concrete implementation of `H1st` AI with you and the open-source data science community!\n\n## Learn the Key Concepts\nHuman-First AI (`H1st` AI) solves three critical challenges in real-world data science:\n\n1. __Industrial AI needs human insight:__ In so many important applications, there isn‘t enough data for ML. For example, last year‘s product‘s data does not apply to this year‘s new model. Or, equipment not yet shipped obviously have no data history to speak of. `H1st` combines human knowledge and any available data to enable intelligent systems, and companies can achieve earlier time-to-market.\n\n2. __Data scientists need human tools:__ Today‘s tools are to compete rather than to collaborate. When multiple data scientists work on the same project, they are effectively competing to see who can build the better model. `H1st` breaks a large modeling problem into smaller, easier parts. This allows true collaboration and high productivity, in ways similar to well-established software engineering methodology. \n\n3. __AI needs human trust:__ AI models can\'t be deployed when they lack user trust. AI increasingly face regulatory challenges. `H1st` supports model description and explanation at multiple layers, enabling transparent and trustworthy AI.\n\n\n## Get started\n`H1st` runs on Python 3.8 or above. Install with \n```\npip install --upgrade pip\npip3 install h1st\n```\nFor Windows, please use 64bit version and install [VS Build Tools](https://visualstudio.microsoft.com/downloads/) before installing H1st.\n\nStart by reading about our [philosophy](https://h1st.readthedocs.io/en/latest/manifesto/README.html) and [Object Model](https://h1st.readthedocs.io/en/latest/concepts/object-model.html)\n\nSee the [Quick Start](https://h1st.readthedocs.io/en/latest/tutorials/quick-start/README.html) for simple "Hello world" examples of using [H1st rule-based model](https://h1st.readthedocs.io/en/latest/tutorials/quick-start/README.html#rule-based-model) & [H1st ML model](https://h1st.readthedocs.io/en/latest/tutorials/quick-start/README.html#mlmodeler-and-mlmodel) and using [H1st Graph](https://h1st.readthedocs.io/en/latest/tutorials/quick-start/README.html#h1st-graph).\n\n\n## Read the Documentation, Tutorials, and API Documentation\n\nGo over [the Concepts](https://h1st.readthedocs.io/en/latest/concepts/README.html)\n\nFor a simple real-world data science example using H1st Modeler and Model API, take a look at\n- [Modeler and Model with Iris dataset](https://h1st.readthedocs.io/en/latest/tutorials/examples/modeler-model.html).\n- [H1st Oracle: Combine Encoded Domain Knowledge with Machine Learning](https://h1st.readthedocs.io/en/latest/tutorials/examples/oracle-iot.html) in which we used Microsoft Azure Predictive Maintenance dataset to demonstrate the power of the Oracle.\n\nTo fully understand H1st philosophy and power, check out the [Use-case examples](https://h1st.readthedocs.io/en/latest/tutorials/use-cases/README.html).\n\nFor a deep dive into the components, please refer to our full [API Documentation](https://h1st.readthedocs.io/en/latest/api/README.html).\n\n## Join and Learn from Our Open-Source Community\nWe are collaborating with the open-source community. For Arimo-Panasonic, use cases include industrial applications such as Cybersecurity, Predictive Maintenance, Fault Prediction, Home Automation, Avionic & Automotive Experience Management, etc.\n\nWe\'d love to see your use cases and your contributions to open-source `H1st` AI. \n',
    'author': 'h1st-ai',
    'author_email': 'engineering@aitomatic.com',
    'maintainer': None,
    'maintainer_email': None,
    'url': 'https://h1st.ai',
    'packages': packages,
    'package_data': package_data,
    'install_requires': install_requires,
    'python_requires': '>=3.8,<3.10',
}


setup(**setup_kwargs)



from typing import Any, NoReturn
import cloudpickle
import s3fs
from h1st.model.repository.storage.base import Storage


class S3Storage(Storage):
    """
    Provide data storage on top of AWS S3
    """

    def __init__(self, bucket_name: str = "", prefix: str = ""):
        """
        :param bucket_name: s3 bucket name to store data into
        :param prefix: s3 object prefix, leave blank to store at bucket root
        """
        self.bucket_name = bucket_name
        self.prefix = prefix
        self.fs = s3fs.S3FileSystem()

    def get_obj(self, name: str) -> Any:
        """
        Retrieve object value

        :param name: object name
        """
        key = self._to_key(name)
        try:
            with self.fs.open(key, 'rb') as f:
                return cloudpickle.load(f)
        except FileNotFoundError as ex:
            raise KeyError(name) from ex

    def get_bytes(self, name) -> bytes:
        """
        Retrieve object value in bytes

        :param name: object name
        """
        key = self._to_key(name)
        try:
            with self.fs.open(key, 'rb') as f:
                return f.read()
        except FileNotFoundError as ex:
            raise KeyError(name) from ex

    def set_obj(self, name: str, value: Any) -> NoReturn:
        """
        Set key value to a python object

        :param name: object name
        :param value: value in python object
        """
        key = self._to_key(name)

        with self.fs.open(key, 'wb') as f:
            return cloudpickle.dump(value, f)

    def set_bytes(self, name: str, value: bytes) -> NoReturn:
        """
        Set a key value to a list of bytes

        :param name: object name
        :param value: value in bytes
        """
        key = self._to_key(name)

        with self.fs.open(key, 'wb') as f:
            f.write(value)

    def exists(self, name: str) -> bool:
        """
        Return true if object exists in the storage
        """
        key = self._to_key(name)
        return self.fs.exists(key)

    def delete(self, name: str) -> NoReturn:
        """
        Delete an object in storage
        """
        try:
            key = self._to_key(name)
            self.fs.rm(key)
        except FileNotFoundError:
            pass

    def _to_key(self, key):
        """
        Convert a key to s3 object key with bucket and prefix
        """
        key = key.replace("/", "_").replace("::", "/")

        if self.prefix:
            key = f"{self.prefix}/{key}"

        return f"{self.bucket_name}/{key}"



from typing import List
from h1st.model.predictive_model import PredictiveModel
from h1st.model.modeler import Modeler
from h1st.model.ml_modeler import MLModeler

class EnsembleModeler(Modeler):
    pass



from h1st.h1flow.ui.has_web_ui import HasWebUI


# __TODO__ remove this file

class H1Step():
    def __init__(self):
        pass

    def execute(self):
        raise NotImplementedError('Please implement this method')


class H1StepWithWebUI(H1Step, HasWebUI):
    def __init__(self):
        pass



from typing import Dict

import numpy as np
import pandas as pd

from h1st.model.predictive_model import PredictiveModel
from h1st.model.rule_based_modeler import RuleBasedModeler

class Ensemble(PredictiveModel):
    """
    Ensembler Model in Segmented World Ensemble framework
    """
    def predict(self, input_data: Dict) -> Dict:
        return {'predictions': None}


class MajorityVotingEnsemble(PredictiveModel):
    """
    Ensemble Model in Oracle framework
    """
    def vote_on(self, df_sub_model_predictions):
        '''
        Do majority voting on sub models' prediction to get the final classificiation result

        :param df_sub_model_predictions: Pandas DataFrame including sum model's prediction in each column
        :returns: a dictionary with key `predictions` and value containing the predictions in numpy array
        '''
        if df_sub_model_predictions.shape[1] == 1:
            predictions = df_sub_model_predictions.iloc[:, 0]
        else:
            predictions = df_sub_model_predictions.mode(axis='columns', numeric_only=True)[0]
            
        return predictions

    def predict(self, input_data: Dict) -> Dict:
        sub_model_predictions = {}
        for name, data in input_data.items():
            # create a fixed length array with max index val
            pred_w_index = np.empty(data['num_of_data']) 
            pred_w_index[:] = np.nan    

            # put predictions in indices
            np.put(pred_w_index, data['index'], data['predictions'])
            sub_model_predictions[name] = pred_w_index
        df_sub_model_predictions = pd.DataFrame(sub_model_predictions)

        return {
            'predictions': self.vote_on(df_sub_model_predictions)
        }


from .kswe import KSWE
from .kswe_modeler import KSWEModeler
from .segmentor import CombinationSegmentor, MaxSegmentationModeler
from .ensemble import Ensemble, MajorityVotingEnsemble
from .sub_model_modeler import RandomForestClassifierModeler